package koropeta.core;
import koropeta.*;
import java.awt.image.*;
import java.awt.color.*;
import java.io.*;
import java.util.*;

public class AlphaSplittedImageReader extends AlphaSplittedImage {
    static AlphaSplittedImageCache cache = new AlphaSplittedImageCache();
    
    public static BufferedImage read(File dir, String basename) throws IOException {
	File pngfile = new File(dir, basename+".png");
	File pnafile = new File(dir, basename+".pna");

	if (!pngfile.isFile()) {
	    throw new RuntimeException("Failed to read: "+pngfile.toString());
	}

	BufferedImage cached = cache.cache(dir, basename);
	if (cached != null) {
	    Main.log("This image is in a cache: "+dir+"/"+basename+".png (and .pna if exists)");
	    return cached;
	}

	Main.log("Attempt to load image: "+pngfile.getAbsolutePath());
	long t = System.currentTimeMillis();

	BufferedImage png = cache.raw_read(pngfile);
	BufferedImage result = null;
	
	if (pnafile.isFile()) {
	    BufferedImage pna = cache.raw_read(pnafile);
	    
	    if (png.getWidth() != pna.getWidth() ||
		png.getHeight() != pna.getHeight()) {

		Main.error(
		    "Image size differs: ["+pngfile+"] v.s. ["+pnafile+"]");
	    }
	    else {
		BufferedImage base = make_32bit_argb(png, false);
		BufferedImage alpha = make_32bit_argb(pna, false);
		
		apply_alpha(base, alpha);
		result = base;
	    }
	}
	else {
	    result = make_32bit_argb(png, true);
	}

	Main.log("... took "+(System.currentTimeMillis() - t)+" milliseconds.");
	return cache.cache(dir, basename, result);
    }

    protected static void apply_alpha(BufferedImage base_img, BufferedImage alpha_img) {
	// base, alphaの何れもTYPE_INT_ARGBで作られたBufferedImageでなければならない。
	int[] base_data  = ((DataBufferInt)(base_img.getRaster().getDataBuffer())).getData();
	int[] alpha_data = ((DataBufferInt)(alpha_img.getRaster().getDataBuffer())).getData();

	for (int i = 0; i < base_data.length; i++) {
	    final int base  = base_data[i];
	    final int alpha = alpha_data[i];

	    base_data[i] =
		(alpha_data[i] & 0x000000ff) << 24 |
		 base_data[i]  & 0x00ffffff;
	}
    }

    protected static BufferedImage make_32bit_argb(BufferedImage src, boolean topleft_mask) {
	BufferedImage dst = new BufferedImage(
	    src.getWidth(), src.getHeight(), BufferedImage.TYPE_INT_ARGB);
	// TYPE_INT_ARGBで作られたBufferedImageは常に次の形式となっている。
	// ColorModel : DirectColorModel (ARGB 32bit)
	// SampleModel: SinglePixelPackedSampleModel
	// DataBuffer : DataBufferInt
	    
	ColorModel src_cmodel = src.getColorModel();
	SampleModel src_smodel = src.getSampleModel();
	DataBuffer src_dbuf = src.getRaster().getDataBuffer();

	ColorModel dst_cmodel = dst.getColorModel();
	SampleModel dst_smodel = dst.getSampleModel();
	DataBuffer dst_dbuf = dst.getRaster().getDataBuffer();

	try {
	    if (false) {}
	    
	    else if (load_8bits_index_onto_direct_argb(
			 src_cmodel, src_smodel, src_dbuf,
			 dst_cmodel, dst_smodel, dst_dbuf,
			 dst, src, topleft_mask)) {}
	    
	    else if (load_3bytes_rgb_onto_direct_argb(
			 src_cmodel, src_smodel, src_dbuf,
			 dst_cmodel, dst_smodel, dst_dbuf,
			 dst, src, topleft_mask)) {}
	    
	    else if (load_nbits_index_onto_direct_argb(
			 src_cmodel, src_smodel, src_dbuf,
			 dst_cmodel, dst_smodel, dst_dbuf,
			 dst, src, topleft_mask)) {}
	    
	    else if (fallback_load_any_onto_direct_argb(
			 src_cmodel, src_smodel, src_dbuf,
			 dst_cmodel, dst_smodel, dst_dbuf,
			 dst, src, topleft_mask)) {}
	    
	    else {
		fallback_load(dst, src, topleft_mask);
	    }
	}
	catch (Exception e) {
	    e.printStackTrace();
	    
	    fallback_load(dst, src, topleft_mask);
	}

	return dst;
    }

    protected static boolean load_8bits_index_onto_direct_argb(
	ColorModel src_cmodel, SampleModel src_smodel, DataBuffer src_dbuf,
	ColorModel dst_cmodel, SampleModel dst_smodel, DataBuffer dst_dbuf,
	BufferedImage newimage, BufferedImage png, boolean topleft_mask) {

	if (!(src_cmodel instanceof IndexColorModel) ||
	    !(src_cmodel.getColorSpace().getType() == ColorSpace.TYPE_RGB) ||
	    !(src_cmodel.getPixelSize() == 8) ||
	    
	    !(src_dbuf instanceof DataBufferByte) ||
	    
	    !(src_smodel.getNumBands() == 1) ||
	    
	    !(dst_cmodel instanceof DirectColorModel) ||
	    !(dst_smodel instanceof SinglePixelPackedSampleModel) ||
	    !(dst_dbuf instanceof DataBufferInt)) {
	    return false;
	}
	
	Main.log("Optimize: src = Index 8bit :: dest = Direct ARGB 32bit");

	final byte[] src = ((DataBufferByte)src_dbuf).getData();
	int[] dest = ((DataBufferInt)dst_dbuf).getData();
	if (src.length != dest.length) {
	    throw new RuntimeException(
		"src.length("+src.length+") != dest.length("+dest.length+")... don't you know why?");
	}

	final byte bg_pix = src[0];

	final int[] map = new int[((IndexColorModel)src_cmodel).getMapSize()];
	((IndexColorModel)src_cmodel).getRGBs(map);

	for (int i = 0; i < src.length; i++) {
	    final byte src_pix = src[i];
	    
	    if (topleft_mask && src_pix == bg_pix) {
		// 透明色。ピクセルを轉冩せず次へ
	    }
	    else {
		dest[i] = map[src_pix & 0xff];
	    }
	}

	return true;
    }

    protected static boolean load_3bytes_rgb_onto_direct_argb(
	ColorModel src_cmodel, SampleModel src_smodel, DataBuffer src_dbuf,
	ColorModel dst_cmodel, SampleModel dst_smodel, DataBuffer dst_dbuf,
	BufferedImage newimage, BufferedImage png, boolean topleft_mask) {

	if (!(src_cmodel instanceof ComponentColorModel) ||
	    !(src_cmodel.getColorSpace().getType() == ColorSpace.TYPE_RGB) ||
	    !(src_cmodel.getPixelSize() == 24) ||
	    
	    !(src_smodel instanceof PixelInterleavedSampleModel) ||
	    !(src_smodel.getNumBands() == 3) ||
	    
	    !(src_dbuf instanceof DataBufferByte) ||
	    
	    !(dst_cmodel instanceof DirectColorModel) ||
	    !(dst_smodel instanceof SinglePixelPackedSampleModel) ||
	    !(dst_dbuf instanceof DataBufferInt)) {
	    return false;
	}
	Main.log("Optimize: src = RGB 24bit :: dest = Direct ARGB 32bit");

	byte[] src = ((DataBufferByte)src_dbuf).getData();
	int[] dest = ((DataBufferInt)dst_dbuf).getData();
	if (src.length != dest.length * 3) {
	    throw new RuntimeException(
		"src.length("+src.length+") != dest.length("+dest.length+") * 3... don't you know why?");
	}

	final byte bg_r = src[0];
	final byte bg_g = src[1];
	final byte bg_b = src[2];

	for (int i = 0, j = 0; i < src.length; i += 3, j++) {
	    final byte src_r = src[i  ];
	    final byte src_g = src[i+1];
	    final byte src_b = src[i+2];

	    if (topleft_mask &&
		src_r == bg_r &&
		src_g == bg_g &&
		src_b == bg_b) {
		// 透明色。ピクセルを轉冩せず次へ
	    }
	    else {
		final int argb =
		    255            << 24 |
		    (src_r & 0xff) << 16 |
		    (src_g & 0xff) <<  8 |
		    (src_b & 0xff);
		
		dest[j] = argb;
	    }
	}

	return true;
    }

    protected static boolean load_nbits_index_onto_direct_argb(
	ColorModel src_cmodel, SampleModel src_smodel, DataBuffer src_dbuf,
	ColorModel dst_cmodel, SampleModel dst_smodel, DataBuffer dst_dbuf,
	BufferedImage newimage, BufferedImage png, boolean topleft_mask) {

	if (!(src_cmodel instanceof IndexColorModel) ||
	    !(src_cmodel.getColorSpace().getType() == ColorSpace.TYPE_RGB) ||
	    !(src_cmodel.getPixelSize() <= 8) ||
	    
	    !(src_dbuf instanceof DataBufferByte) ||

	    !(src_smodel instanceof MultiPixelPackedSampleModel) ||
	    !(src_smodel.getNumBands() == 1) ||
	    !(((MultiPixelPackedSampleModel)src_smodel).getDataBitOffset() == 0) ||
	    
	    !(dst_cmodel instanceof DirectColorModel) ||
	    !(dst_smodel instanceof SinglePixelPackedSampleModel) ||
	    !(dst_dbuf instanceof DataBufferInt)) {
	    return false;
	}
	
	final int nbits = src_cmodel.getPixelSize();
	final int bitmask = (1 << nbits) - 1;
	
	Main.log("Optimize: src = Index 1bit :: dest = Direct ARGB 32bit");

	final byte[] src = ((DataBufferByte)src_dbuf).getData();
	int[] dest = ((DataBufferInt)dst_dbuf).getData();

	final int bg_pix = (src[0] >> (8 - nbits)) & bitmask;

	final int[] map = new int[((IndexColorModel)src_cmodel).getMapSize()];
	((IndexColorModel)src_cmodel).getRGBs(map);

	final int width = png.getWidth();
	final int height = png.getHeight();
	final int scanline_stride = ((MultiPixelPackedSampleModel)src_smodel).getScanlineStride();
	
	for (int y = 0, row_start = 0, dest_i = 0;
	     y < height;
	     y++, row_start += scanline_stride) {
	    
	    for (int x = 0, i = row_start;
		 x < width;
		 i++) {
		
		// 1バイトにつき(8 / nbits)ピクセルが表現されてゐる。
		for (int j = 8 - nbits;
		     j >= 0 && x < width;
		     j -= nbits, x++, dest_i++) {
		    
		    final int src_pix = (src[i] >> j) & bitmask;

		    if (topleft_mask && src_pix == bg_pix) {
			// 透明色。ピクセルを轉冩せず次へ
		    }
		    else {
			dest[dest_i] = map[src_pix];
		    }
		}
	    }
	}

	return true;
    }

    protected static boolean fallback_load_any_onto_direct_argb(
	ColorModel src_cmodel, SampleModel src_smodel, DataBuffer src_dbuf,
	ColorModel dst_cmodel, SampleModel dst_smodel, DataBuffer dst_dbuf,
	BufferedImage newimage, BufferedImage png, boolean topleft_mask) {

	if (!(dst_cmodel instanceof DirectColorModel) ||
	    !(dst_dbuf instanceof DataBufferInt)) {
	    return false;
	}
	Main.log("Optimize: src = ANY :: dest = Direct ARGB 32bit\n"+
		 "          "+src_dbuf.getClass()+"\n"+
		 "          +-- "+src_dbuf);
	Main.log(describe_color_model(src_cmodel));
	Main.log(describe_sample_model(src_smodel));

	final Object bg_pixel = png.getRaster().getDataElements(0, 0, null);
	//final int[] bg_compo = png.getColorModel().getComponents(bg_pixel, null, 0);
	final int bg_compo = png.getColorModel().getRGB(bg_pixel);
	final int bg_r = (bg_compo >> 24) & 0xff;
	final int bg_g = (bg_compo >> 16) & 0xff;
	final int bg_b = (bg_compo >>  8) & 0xff;

	Object src_pixel = png.getRaster().getDataElements(0, 0, null);

	int[] dest = ((DataBufferInt)dst_dbuf).getData();
	int dest_i = 0;

	final int width = png.getWidth();
	final int height = png.getHeight();
	for (int y = 0; y < height; y++) {
	    for (int x = 0; x < width; x++) {
		png.getRaster().getDataElements(x, y, src_pixel);
		final int src_compo = png.getColorModel().getRGB(src_pixel);
		final int src_r = (src_compo >> 24) & 0xff;
		final int src_g = (src_compo >> 16) & 0xff;
		final int src_b =  src_compo        & 0xff;
		
		if (topleft_mask &&
		    src_compo == bg_compo) {
		    // 透明色。ピクセルを轉冩せず次へ
		}
		else {
		    final int argb =
			255          << 24 |
			src_r        << 16 |
			src_g        <<  8 |
			src_b;

		    dest[dest_i] = argb;
		}

		dest_i++;
	    }
	}

	return true;
    }

    protected static void fallback_load(BufferedImage newimage, BufferedImage png, boolean topleft_mask) {
	Main.log("GOD DAMN! All our attempts to optimize has been failed. Using the final fallback...\n"+
		 "    "+png.getRaster().getDataBuffer().getClass());
	Main.log(describe_color_model(png.getColorModel()));
	Main.log(describe_sample_model(png.getSampleModel()));
	
	Object bg_pixel = png.getRaster().getDataElements(0, 0, null);
	final int bg_compo = png.getColorModel().getRGB(bg_pixel);
	final int bg_r = (bg_compo >> 24) & 0xff;
	final int bg_g = (bg_compo >> 16) & 0xff;
	final int bg_b = (bg_compo >>  8) & 0xff;

	Object src_pixel = png.getRaster().getDataElements(0, 0, null);

	Object dst_pixel = newimage.getRaster().getDataElements(0, 0, null);
	int[] dst_compo = newimage.getColorModel().getComponents(dst_pixel, null, 0);

	final int width = png.getWidth();
	final int height = png.getHeight();
	for (int y = 0; y < height; y++) {
	    for (int x = 0; x < width; x++) {
		png.getRaster().getDataElements(x, y, src_pixel);
		final int src_compo = png.getColorModel().getRGB(src_pixel);
		final int src_r = (src_compo >> 24) & 0xff;
		final int src_g = (src_compo >> 16) & 0xff;
		final int src_b =  src_compo        & 0xff;
		
		if (topleft_mask &&
		    src_compo == bg_compo) {
		    // 透明色。ピクセルを轉冩せず次へ
		}
		else {
		    dst_compo[0] = src_r;
		    dst_compo[1] = src_g;
		    dst_compo[2] = src_b;
		    dst_compo[3] = 255;
		    
		    newimage.getColorModel().getDataElements(dst_compo, 0, dst_pixel);
		    newimage.getRaster().setDataElements(x, y, dst_pixel);
		}
	    }
	}
    }
}