package koropeta.core;
import java.util.*;
import java.util.regex.*;
import java.io.*;
import java.nio.charset.*;

public class BlockedDescription extends TreeMap {
    static final Pattern re_comment = Pattern.compile("^//");
    static final Pattern re_pair = Pattern.compile("^(.+?)\\s*,\\s*(.+)$");

    String charset = "MS932";
    boolean is_dirty = false;

    public BlockedDescription() {
	this(null);
    }

    public BlockedDescription(File f) {
	super(new NumericComparator());
	
	if (f == null) {
	    return;
	}

	try {
	    new Parser().parse(new FileInputStream(f));
	}
	catch (FileNotFoundException e) {
	    // 無視
	}
	catch (Exception e) {
	    e.printStackTrace();
	}
    }

    protected boolean cp932_encodable_p() throws Exception {
	/*
	  現在の内容を完全にCP932にエンコード出來る場合は眞を返す。
	  このメソッドが僞を返す場合、ファイルへの保存はUTF-8で
	  行はれなければならない。
	*/
	Charset cset = Charset.forName("MS932");
	CharsetEncoder enc = cset.newEncoder();
	
	for (Iterator ite_key = keySet().iterator(); ite_key.hasNext(); ) {
	    String key = (String)ite_key.next();

	    if (!enc.canEncode(key)) {
		return false;
	    }

	    Object value = get(key);
	    if (value instanceof String) {
		if (!enc.canEncode((String)value)) {
		    return false;
		}
	    }
	    else if (value instanceof Map) {
		Map map = (Map)value;

		for (Iterator ite = map.keySet().iterator(); ite.hasNext(); ) {
		    String k = (String)ite.next();
		    String v = (String)map.get(k);

		    if (!enc.canEncode(k) || !enc.canEncode(v)) {
			return false;
		    }
		}
	    }
	    else {
		throw new Exception("Illegal value for key: "+key+": "+value);
	    }
	}

	return true;
    }

    public void save(File f) {
	try {
	    save(new FileOutputStream(f));
	}
	catch (Exception e) {
	    throw new RuntimeException(e);
	}
    }
    public void save(OutputStream os) throws Exception {
	charset = cp932_encodable_p() ? "MS932" : "UTF-8";
	String crlf = new String(new char[] {0x0d, 0x0a});

	BufferedWriter writer = new BufferedWriter(
	    new OutputStreamWriter(os, charset));
	writer.write("charset,");
	writer.write(charset.equals("MS932") ? "Shift_JIS" : charset);
	writer.write(crlf);

	for (Iterator ite_key = keySet().iterator(); ite_key.hasNext(); ) {
	    String key = (String)ite_key.next();
	    Object value = get(key);

	    if (key.equals("charset")) {
		continue;
	    }

	    if (value instanceof String) {
		writer.write(crlf);
		writer.write(key);
		writer.write(',');
		writer.write((String)value);
		writer.write(crlf);
	    }
	    else if (value instanceof Map) {
		Map map = (Map)value;

		if (map.size() == 0) {
		    // 空のマップは保存しない。
		    continue;
		}

		writer.write(crlf);
		writer.write(key);
		writer.write(crlf);
		writer.write('{');
		writer.write(crlf);
		
		for (Iterator ite = map.keySet().iterator(); ite.hasNext(); ) {
		    String k = (String)ite.next();
		    String v = (String)map.get(k);
		    
		    writer.write(k);
		    writer.write(',');
		    writer.write(v);
		    writer.write(crlf);
		}

		writer.write('}');
		writer.write(crlf);
	    }
	    else {
		throw new Exception("Illegal value for key: "+key+": "+value);
	    }
	}

	writer.close();
    }

    public final boolean dirty() {
	return is_dirty;
    }
    public final boolean dirty(boolean b) {
	is_dirty = b;
	return b;
    }

    private class Parser {
	BufferedReader reader;
	Vector overread;

	public Parser() {
	    overread = new Vector();
	}

	private String next() throws Exception {
	    String line;
	    if (overread.size() > 0) {
		line = (String)overread.get(0);
		overread.remove(0);
	    }
	    else {
		line = reader.readLine();

		if (line != null) {
		    line = line.trim();
		}
	    }
	    return line;
	}

	private void pushback(String line) {
	    if (line != null) {
		overread.add(line);
	    }
	}
	
	public void parse(InputStream is) throws Exception {
	    // 最初の行を讀んでみて、それがcharset指定だったら讀み直す。
	    is.mark(1024);
	    
	    reader = new BufferedReader(
		new InputStreamReader(is, charset));

	    Matcher m;
	    String first_line = next();
	    pushback(first_line);
	    
	    if ((m = re_pair.matcher(first_line)).find()) {
		if (m.group(1).equals("charset")) {
		    String new_charset = m.group(2);

		    // CP932 / Shift_JIS / SJISを指定されたらMS932を使ふ。
		    if (new_charset.equals("Shift_JIS") ||
			new_charset.equals("SJIS") ||
			new_charset.equals("CP932")) {
			
			new_charset = "MS932";
		    }

		    if (!charset.equals(new_charset)) {
			charset = new_charset;
			is.reset();

			reader = new BufferedReader(
			    new InputStreamReader(is, charset));
		    }
		}
	    }

	    while (true) {
		String line = next();
		if (line == null) {
		    break;
		}

		List keys = new Vector();
		
		if (line.length() == 0) {
		    continue;
		}
		else if ((m = re_comment.matcher(line)).find()) {
		    continue;
		}
		else if ((m = re_pair.matcher(line)).find()) {
		    // 次の行が「{」だったらエントリ複製。
		    // さうでなければ單一行エントリ。
		    String next = next();
		    pushback(next);

		    if (next != null && next.trim().equals("{")) {
			StringTokenizer st_keys = new StringTokenizer(line, ",");
			while (st_keys.hasMoreTokens()) {
			    keys.add(st_keys.nextToken().trim());
			}
		    }
		    else {
			put(m.group(1), m.group(2));
			continue;
		    }
		}
		else {
		    keys.add(line);
		}

		Map value = new TreeMap(new NumericComparator());
		String brace = next();
		if (brace == null || !brace.equals("{")) {
		    throw new Exception("Syntax error: `{' not found after `"+line+"': `"+brace+"'");
		}
		
		while (true) {
		    String inner_line = next();
		    if (inner_line == null) {
			throw new Exception("Syntax error: `}' not found before EOF");
		    }
		    else if (inner_line.length() == 0) {
			continue;
		    }
		    else if ((m = re_comment.matcher(inner_line)).find()) {
			continue;
		    }
		    else if (inner_line.equals("}")) {
			break;
		    }
		    else if ((m = re_pair.matcher(inner_line)).find()) {
			value.put(m.group(1), m.group(2));
		    }
		}

		for (Iterator ite = keys.iterator(); ite.hasNext(); ) {
		    put(ite.next(), value);
		}
	    }

	    reader.close();
	}
    }

    public TreeMap putMap(String key) {
	TreeMap map = new TreeMap(new NumericComparator());
	put(key, map);
	return map;
    }
}