package koropeta.core;
import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import javax.swing.*;
import javax.swing.event.*;
import java.util.*;
import java.lang.ref.*;

public class LayeredImage implements Cloneable {
    java.util.List layers; // Layer
    Dimension size;
    boolean is_dirty;

    SoftReference composed; // BufferedImage
    ThumbnailMaker thumb_maker;
    Vector list_data_listeners;
    
    public LayeredImage() {
	layers = Collections.synchronizedList(new Vector());
	size = new Dimension(0, 0);
	thumb_maker = new ThumbnailMaker();
	list_data_listeners = new Vector();
    }

    public Object clone() {
	LayeredImage clo = new LayeredImage();
	
	for (Iterator ite = layers.iterator(); ite.hasNext(); ) {
	    Layer l = (Layer)ite.next();
	    clo.layers.add(l.clone());
	}

	clo.size = new Dimension(size);
	clo.is_dirty = is_dirty;
	return clo;
    }

    public Layer add(String name, BufferedImage image) {
	return add(name, image, 0, 0, true);
    }
    public Layer add(String name, BufferedImage image, int x, int y, boolean visible) {
	Layer layer = new RealLayer(name, image);
	layer.x(x);
	layer.y(y);
	layer.visible(visible);
	synchronized (layers) {
	    layers.add(layer);
	}

	/*if (layers.size() == 1) {
	    // 空のイメージ => 空でないイメージへの變化
	    // 統合後サイズを設定。
	    size = new Dimension(image.getWidth(), image.getHeight());
	    }*/
	if (size.width < x + image.getWidth() ||
	    size.height < y + image.getHeight()) {
	    size = new Dimension(
		(int)(Math.max(size.width, x + image.getWidth())),
		(int)(Math.max(size.height, y + image.getHeight())));
	}

	changed(ListDataEvent.INTERVAL_ADDED, layers.size() - 1);
	return layer;
    }

    public Layer add(String name, LayeredImage image) {
	return add(name, image, 0, 0, true);
    }
    public Layer add(String name, LayeredImage image, int x, int y, boolean visible) {
	Layer layer = new ReferenceLayer(name, image);
	layer.x(x);
	layer.y(y);
	layer.visible(visible);
	synchronized (layers) {
	    layers.add(layer);
	}

	/*if (layers.size() == 1) {
	    // 空のイメージ => 空でないイメージへの變化
	    // 統合後サイズを設定。
	    size = image.size();
	    }*/
	if (size.width < x + image.getWidth() ||
	    size.height < y + image.getHeight()) {
	    size = new Dimension(
		(int)(Math.max(size.width, x + image.getWidth())),
		(int)(Math.max(size.height, y + image.getHeight())));
	}

	changed(ListDataEvent.INTERVAL_ADDED, layers.size() - 1);
	return layer;
    }

    public final boolean dirty() {
	return is_dirty;
    }
    public final boolean dirty(boolean bool) {
	is_dirty = bool;
	return bool;
    }

    public void insert(int idx, Layer layer) {
	synchronized (layers) {
	    layers.add(idx, layer);
	}
	changed(ListDataEvent.INTERVAL_ADDED, idx);
    }

    public void remove(int idx) {
	Layer l = (Layer)layers.get(idx);
	l.close();

	synchronized (layers) {
	    layers.remove(idx);
	}
	
	changed(ListDataEvent.INTERVAL_REMOVED, idx);
    }

    public void remove(String name) {
	synchronized (layers) {
	    int idx = indexOf(name);
	    if (idx != -1) {
		remove(idx);
	    }
	}
    }

    public void clear() {
	int old_size = layers.size();
	if (old_size > 0) {
	    synchronized (layers) {
		for (Iterator ite = layers.iterator(); ite.hasNext(); ) {
		    Layer l = (Layer)ite.next();
		    l.close();
		}
		
		layers.clear();
	    }

	    changed(ListDataEvent.INTERVAL_REMOVED, 0, old_size-1);
	}
    }

    public Dimension size() {
	return new Dimension(size);
    }
    public Dimension size(Dimension newsize) {
	if (newsize.width != size.width ||
	    newsize.height != size.height) {
	    
	    size = new Dimension(newsize);
	    changed(ListDataEvent.CONTENTS_CHANGED, 0, layers.size() - 1);
	}
	return size();
    }
    public final int getWidth() {
	return size.width;
    }
    public final int getHeight() {
	return size.height;
    }

    public Iterator iterator() {
	return layers.iterator();
    }

    public java.util.List layers() {
	// 同期化を崩してはならない爲、Collections.unmodifiableList()は使へないが、
	// 返されるリストを變更してはならない。
	return layers;
    }

    public synchronized BufferedImage compose() {
	if (composed == null || composed.get() == null) {
	    if (layers.isEmpty() || size.width == 0 || size.height == 0) {
		return null;
	    }
	    else {
		BufferedImage bi = new BufferedImage(
		    size.width, size.height, BufferedImage.TYPE_INT_ARGB);
		composed = new SoftReference(bi);

		Graphics2D g2 = bi.createGraphics();
		for (Iterator ite = layers.iterator(); ite.hasNext();) {
		    Layer layer = (Layer)ite.next();
		    
		    if (layer.visible()) {
			BufferedImage img = layer.get();
			if (img != null) {
			    g2.drawImage(img, null, layer.x(), layer.y());
			}
		    }
		}
	    }
	}
	
	return (BufferedImage)composed.get();
    }

    public BufferedImage thumbnail(Dimension dim) {
	return thumbnail(dim.width, dim.height);
    }
    public synchronized BufferedImage thumbnail(int width, int height) {
	thumb_maker.expire_cache();
	return thumb_maker.make(compose(), width, height);
    }
    
    public ImageIcon thumbnail_icon(Dimension dim) {
	return thumbnail_icon(dim.width, dim.height);
    }
    public synchronized ImageIcon thumbnail_icon(int width, int height) {
	thumb_maker.expire_cache();
	return thumb_maker.make_icon(compose(), width, height);
    }

    public final int num_layers() {
	return layers.size();
    }

    public int indexOf(String name) {
	for (int i = 0; i < layers.size(); i++) {
	    Layer layer = (Layer)layers.get(i);
	    
	    if (layer.name().equals(name)) {
		return i;
	    }
	}
	return -1;
    }

    public final Layer layer(int index) {
	return (Layer)layers.get(index);
    }
    
    public final Layer layer(String name) {
	// 名前がnameに等しいレイヤーが存在するなら返す。
	int idx = indexOf(name);
	return idx == -1 ? null : layer(idx);
    }

    public final boolean exists(String name) {
	// 名前がnameに等しいレイヤーが存在するならtrue
	return layer(name) != null;
    }

    public void swap(int id0, int id1) {
	synchronized (layers) {
	    Layer l = (Layer)layers.get(id0);
	    layers.set(id0, layers.get(id1));
	    layers.set(id1, l);
	}

	if (id0 <= id1) {
	    changed(ListDataEvent.CONTENTS_CHANGED, id0, id1);
	}
	else {
	    changed(ListDataEvent.CONTENTS_CHANGED, id1, id0);
	}
    }

    public boolean simple_p() {
	/*
	  以下の全ての條件が成立してゐればtrue

	  * レイヤが1枚のみ
	  * レイヤが非表示でない
	  * レイヤの位置が (0, 0) である
	  * レイヤと全體の大きさが一致してゐる
	*/
	if (layers.size() != 1) {
	    return false;
	}

	Layer l = layer(0);
	if (!l.visible()) {
	    return false;
	}
	if (l.x() != 0 || l.y() != 0) {
	    return false;
	}
	if (size.width != l.get().getWidth() ||
	    size.height != l.get().getHeight()) {
	    return false;
	}

	return true;
    }

    public void addListDataListener(ListDataListener listener) {
	if (!list_data_listeners.contains(listener)) {
	    list_data_listeners.add(listener);
	}
    }

    public void removeListDataListener(ListDataListener listener) {
	list_data_listeners.remove(listener);
    }

    protected void changed(int type, int index) {
	changed(type, index, index);
    }
    protected void changed(int type, int index0, int index1) {
	composed = null;
	
	ListDataEvent ev = new ListDataEvent(
	    this, type, index0, index1);
	
	for (Iterator ite = list_data_listeners.iterator(); ite.hasNext();) {
	    ListDataListener listener = (ListDataListener)ite.next();
	    
	    if (type == ListDataEvent.INTERVAL_ADDED) {
		listener.intervalAdded(ev);
	    }
	    else if (type == ListDataEvent.INTERVAL_REMOVED) {
		listener.intervalRemoved(ev);
	    }
	    else {
		listener.contentsChanged(ev);
	    }
	}
    }

    public interface Layer extends Cloneable {
	public Object clone();
	public String name();
	public BufferedImage get();
	public int x();
	public int x(int newval);
	public int y();
	public int y(int newval);
	public boolean visible();
	public boolean visible(boolean b);
	public BufferedImage thumbnail(Dimension dim);
	public BufferedImage thumbnail(int width, int height);
	public ImageIcon thumbnail_icon(Dimension dim);
	public ImageIcon thumbnail_icon(int width, int height);
	public void close();
    }

    public class ReferenceLayer implements Layer {
	String name;
	LayeredImage ref;
	int x;
	int y;
	boolean visible;

	ThumbnailMaker thumb_maker;
	ListDataListener list_data_listener;

	protected ReferenceLayer(String name, LayeredImage image) {
	    this.name = name;
	    this.ref = image;
	    x = 0;
	    y = 0;
	    visible = true;

	    thumb_maker = new ThumbnailMaker();

	    // このLayeredImageの變更を知りたい。
	    list_data_listener = new ListDataListener() {
		    public void intervalAdded(ListDataEvent e) {
			call_changed();
		    }
		    public void intervalRemoved(ListDataEvent e) {
			call_changed();
		    }
		    public void contentsChanged(ListDataEvent e) {
			call_changed();
		    }

		    public void call_changed() {
			changed(ListDataEvent.CONTENTS_CHANGED, layers.indexOf(this));

			if (size.width < x + ref.size().width ||
			    size.height < y + ref.size().height) {

			    Dimension new_size = new Dimension(
				(int)(Math.max(size.width, x + ref.size().width)),
				(int)(Math.max(size.height, y + ref.size().height)));
			    size(new_size);
			}
		    }
		};
	    ref.addListDataListener(list_data_listener);
	}

	public void close() {
	    ref.removeListDataListener(list_data_listener);
	}

	public LayeredImage layers() {
	    return ref;
	}
	public LayeredImage layers(LayeredImage img) {
	    ref = img;
	    changed(ListDataEvent.CONTENTS_CHANGED, layers.indexOf(this));
	    return img;
	}

	public synchronized Object clone() {
	    ReferenceLayer clone = new ReferenceLayer(name, ref);
	    clone.x = x;
	    clone.y = y;
	    clone.visible = visible;
	    return clone;
	}

	public String name() {
	    return name;
	}

	public BufferedImage get() {
	    return ref.compose();
	}

	public int x() {
	    return x;
	}
	public int x(int newval) {
	    if (x != newval) {
		x = newval;
		
		changed(ListDataEvent.CONTENTS_CHANGED, layers.indexOf(this));
	    }
	    return x;
	}

	public int y() {
	    return y;
	}
	public int y(int newval) {
	    if (y != newval) {
		y = newval;

		changed(ListDataEvent.CONTENTS_CHANGED, layers.indexOf(this));
	    }
	    return y;
	}

	public boolean visible() {
	    return visible;
	}

	public boolean visible(boolean b) {
	    visible = b;

	    changed(ListDataEvent.CONTENTS_CHANGED, layers.indexOf(this));
	    
	    return visible;
	}

	public BufferedImage thumbnail(Dimension dim) {
	    return thumbnail(dim.width, dim.height);
	}
	public synchronized BufferedImage thumbnail(int width, int height) {
	    thumb_maker.expire_cache();
	    return thumb_maker.make(ref.compose(), width, height);
	}

	public ImageIcon thumbnail_icon(Dimension dim) {
	    return thumbnail_icon(dim.width, dim.height);
	}
	public synchronized ImageIcon thumbnail_icon(int width, int height) {
	    thumb_maker.expire_cache();
	    return thumb_maker.make_icon(ref.compose(), width, height);
	}
    }

    public class RealLayer implements Layer {
	String name;
	BufferedImage image;
	int x;
	int y;
	boolean visible;

	ThumbnailMaker thumb_maker;

	protected RealLayer(String name, BufferedImage image) {
	    this.name = name;
	    this.image = image;
	    x = 0;
	    y = 0;
	    visible = true;

	    thumb_maker = new ThumbnailMaker();
	}

	public void close() {
	    // 何もしない。
	}

	public synchronized Object clone() {
	    // FIXME: imageはdeepcopyすべきか？
	    RealLayer clone = new RealLayer(name, image);
	    clone.x = x;
	    clone.y = y;
	    clone.visible = visible;
	    return clone;
	}

	public String name() {
	    return name;
	}

	public BufferedImage get() {
	    return image;
	}

	public int x() {
	    return x;
	}
	public int x(int newval) {
	    if (x != newval) {
		x = newval;
		
		changed(ListDataEvent.CONTENTS_CHANGED, layers.indexOf(this));
	    }
	    return x;
	}

	public int y() {
	    return y;
	}
	public int y(int newval) {
	    if (y != newval) {
		y = newval;

		changed(ListDataEvent.CONTENTS_CHANGED, layers.indexOf(this));
	    }
	    return y;
	}

	public boolean visible() {
	    return visible;
	}

	public boolean visible(boolean b) {
	    visible = b;

	    changed(ListDataEvent.CONTENTS_CHANGED, layers.indexOf(this));
	    
	    return visible;
	}

	public BufferedImage thumbnail(Dimension dim) {
	    return thumbnail(dim.width, dim.height);
	}
	public synchronized BufferedImage thumbnail(int width, int height) {
	    return thumb_maker.make(image, width, height);
	}

	public ImageIcon thumbnail_icon(Dimension dim) {
	    return thumbnail_icon(dim.width, dim.height);
	}
	public synchronized ImageIcon thumbnail_icon(int width, int height) {
	    return thumb_maker.make_icon(image, width, height);
	}
    }
}