package koropeta.core;
import koropeta.*;
import koropeta.core.*;
import java.io.*;
import java.util.*;
import java.util.regex.*;

public class Shell {
    static final Pattern re_surface_file = Pattern.compile(
	"^(surface(\\d+))\\.(png|krps)$", Pattern.CASE_INSENSITIVE);

    static final Pattern re_surface = Pattern.compile(
	"^surface(\\d+)$");
    static final Pattern re_element = Pattern.compile(
	"^element(\\d+)$");
    
    File shelldir;
    Map surfaces; // <Integer, Surface>
    Description descript_txt;
    BlockedDescription surfaces_txt;
    PlainText readme;
    
    public Shell() {
	this(null);
    }

    public Shell(File dir) {
	shelldir = dir;
	surfaces = Collections.synchronizedMap(new TreeMap());
	if (dir == null) {
	    descript_txt = new Description();
	    surfaces_txt = new BlockedDescription();
	    readme = new PlainText();
	}
	else {
	    descript_txt = new Description(new File(dir, "descript.txt"));
	    surfaces_txt = new BlockedDescription(new File(dir, "surfaces.txt"));
	    readme = new PlainText(new File(dir, get_readme_filename()));
	}
	
	load();
    }

    public String find_surface_having_this_id(int id) {
	// ディレクトリ内に、idに一致するsurface0001.pngのやうな畫像が
	// 存在してゐるなら、パディングされたidを返す。
	if (shelldir == null || !shelldir.exists()) {
	    return null;
	}

	String[] files = shelldir.list();
	for (int i = 0; i < files.length; i++) {
	    Matcher m = re_surface_file.matcher(files[i]);
	    if (m.find()) {
		int ident = -1;
		try {
		    ident = Integer.parseInt(m.group(2));
		}
		catch (NumberFormatException e) {}

		if (id == ident) {
		    return m.group(2);
		}
	    }
	}

	return null;
    }

    public synchronized Surface add_clone(Surface source, int id) {
	// 渡されたsurfaceのcloneをこのシェルに登録する。
	String str_id = find_surface_having_this_id(id);
	if (str_id == null) {
	    str_id = Integer.toString(id);
	}

	Surface s = (Surface)source.clone();
	try {
	    s.set_id(str_id);
	}
	catch (NumberFormatException e) {}
	s.dir(shelldir);
	s.basename("surface"+str_id);
	s.endirty();

	surfaces.put(new Integer(id), s);
	return s;
    }

    public synchronized Surface add_surface(int id) {
	// ディレクトリ内に、surface0001.pngのやうにパディングされた畫像が
	// 既に存在してゐるなら、パディングされたものをIDとして使ふ。
	String str_id = find_surface_having_this_id(id);
	if (str_id == null) {
	    // 無いのでパディングせずに文字列化
	    str_id = Integer.toString(id);
	}
	
	Surface s = new Surface(str_id);
	s.dir(shelldir);
	s.basename("surface"+str_id);

	surfaces.put(new Integer(id), s);
	return s;
    }

    public void remove_surface(Surface s) {
	// 指定されたサーフィスに關する情報を抹消する。
	if (!surfaces.containsKey(new Integer(s.get_int_id()))) {
	    throw new RuntimeException(
		"The surface is not in this shell: surface=["+s+"]; shell=["+this+"]");
	}
	surfaces.remove(new Integer(s.get_int_id()));
	surfaces_txt.remove("surface"+s.get_id());
	surfaces_txt.remove("surface"+s.get_int_id());
	surfaces_txt.dirty(true);
	s.remove_file();
    }

    public Map get_surfaces() {
	// 變更不可能なMapを返す。
	return Collections.unmodifiableMap(surfaces);
    }

    public boolean have_surface(int id) {
	return surfaces.containsKey(new Integer(id));
    }

    public Map get_description() {
	// 變更不可能なMapを返す。
	return Collections.unmodifiableMap(descript_txt);
    }

    public Description modifiable_description() {
	return descript_txt;
    }

    public final File shell_directory() {
	return shelldir;
    }

    public File shell_directory(File dir) {
	shelldir = dir;

	for (Iterator ite = surfaces.values().iterator(); ite.hasNext(); ) {
	    Surface s = (Surface)ite.next();
	    s.dir(dir);
	}
	
	return dir;
    }

    public PlainText readme() {
	return readme;
    }

    public String get_readme_filename() {
	return (surfaces_txt.get("readme") != null ?
		(String)surfaces_txt.get("readme") : "readme.txt");
    }

    public boolean dirty() {
	if (descript_txt.dirty()) {
	    return true;
	}

	if (surfaces_txt.dirty()) {
	    return true;
	}

	if (readme.dirty()) {
	    return true;
	}
	
	for (Iterator ite = surfaces.values().iterator(); ite.hasNext(); ) {
	    Surface s = (Surface)ite.next();
	    if (s.dirty()) {
		return true;
	    }
	}
	return false;
    }

    protected Map desc_for_surface(Surface s, boolean make_new) {
	// surface00001, surface1
	// 何れの形式も用ゐられる。
	Map desc = null;
	if (surfaces_txt.containsKey("surface"+s.get_id())) {
	    desc = (Map)surfaces_txt.get("surface"+s.get_id());
	}
	else if (surfaces_txt.containsKey("surface"+s.get_int_id())) {
	    desc = (Map)surfaces_txt.get("surface"+s.get_int_id());
	}
	else {
	    if (make_new) {
		// 新たに追加
		desc = surfaces_txt.putMap("surface"+s.get_id());
	    }
	}
	return desc;
    }

    public int optimum_version() {
	/*
	  prefで次の三択:
	  1. 常にSERIKO/1.x
	  2. なるべくSERIKO/1.x (表現不可能なら2.0)
	  3. 常にSERIKO/2.0
	*/
	String manner = Main.ini().getProperty(
	    "koropeta.pref.serikoVersion", "1-as-possible");
	
	if (manner.equals("always-1")) {
	    return 0;
	}
	else if (manner.equals("1-as-possible")) {
	    for (Iterator ite = surfaces.values().iterator(); ite.hasNext(); ) {
		Surface s = (Surface)ite.next();

		if (s.least_version() == 1) {
		    return 1;
		}
	    }
	    return 0;
	}
	else {
	    return 1;
	}
    }

    public void save() {
	if (shelldir.exists() && !shelldir.isDirectory()) {
	    shelldir.delete();
	}
	if (!shelldir.exists()) {
	    shelldir.mkdir();
	}

	boolean animation_is_dirty = false;
	for (Iterator ite = surfaces.values().iterator(); ite.hasNext(); ) {
	    Surface s = (Surface)ite.next();
	    if (s.dirty()) {
		s.save();

		if (s.collisions().dirty() || s.elements().dirty() || s.animation().dirty()) {
		    Map desc = desc_for_surface(s, true);
		    
		    s.collisions().save(desc);
		    s.save_elements(desc);
		    animation_is_dirty = true;
		    
		    surfaces_txt.dirty(true);
		}
	    }
	}

	if (animation_is_dirty) {
	    // 適切なバージョンでsurfaces.txtを再構築
	    int version = optimum_version();

	    Map descript = (Map)surfaces_txt.get("descript");
	    if (descript == null) {
		descript = surfaces_txt.putMap("descript");
	    }
	    descript.put("version", Integer.toString(version));
	    surfaces_txt.dirty(true);

	    for (Iterator ite = surfaces.values().iterator(); ite.hasNext(); ) {
		Surface s = (Surface)ite.next();
		Map desc = desc_for_surface(s, true);
		
		s.animation().save(desc, version);
		surfaces_txt.dirty(true);
	    }
	}

	if (descript_txt.dirty()) {
	    // typeは常にshell
	    descript_txt.put("type", "shell");
	    
	    descript_txt.save(new File(shelldir, "descript.txt"));
	    descript_txt.dirty(false);
	}

	if (surfaces_txt.dirty()) {
	    surfaces_txt.save(new File(shelldir, "surfaces.txt"));
	    surfaces_txt.dirty(false);
	}

	if (readme.dirty()) {
	    File file = new File(shelldir, get_readme_filename());
	    if (readme.text().length() != 0) {
		readme.save(file);
	    }
	    else {
		file.delete();
	    }
	    readme.dirty(false);
	}
    }

    protected void load() {
	if (shelldir == null) {
	    return;
	}

	// surfaces.txtからelement定義を持ってゐるエントリをスキャン。
	for (Iterator ite = surfaces_txt.keySet().iterator(); ite.hasNext(); ) {
	    String key = (String)ite.next();

	    Matcher m;
	    if ((m = re_surface.matcher(key)).find()) {
		String id = m.group(1);
	    
		if (surfaces_txt.get(key) instanceof Map) {
		    Map map = (Map)surfaces_txt.get(key);
		    
		    for (Iterator ite_key = map.keySet().iterator(); ite_key.hasNext(); ) {
			String k = (String)ite_key.next();
			
			if ((m = re_element.matcher(k)).find()) {
			    // elementが存在する。
			    Surface s = new Surface(id);
			    s.dir(shelldir);
			    s.basename("surface"+id);
			    s.load_elements(map);
			    
			    surfaces.put(new Integer(s.get_int_id()), s);
			}
		    }
		}
	    }
	}

	// ディレクトリ内の surface000.png 若しくは surface000.krps
	// と云ふ形をした全ファイルをスキャン。
	String[] files = shelldir.list();
	for (int i = 0; i < files.length; i++) {
	    Matcher m;
	    if ((m = re_surface_file.matcher(files[i])).find()) {
		Surface s = new Surface(m.group(2));
		s.dir(shelldir);
		s.basename(m.group(1));

		// element定義があればそちらを優先。
		Integer key = new Integer(s.get_int_id());
		if (!surfaces.containsKey(key)) {
		    surfaces.put(key, s);
		}
	    }
	}

	// SERIKOバージョンを知る
	int version = seriko_version();

	// collisionやanimationを讀む
	for (Iterator ite = surfaces.values().iterator(); ite.hasNext(); ) {
	    Surface s = (Surface)ite.next();
	    Map desc = desc_for_surface(s, false);

	    if (desc != null) {
		s.collisions().load(desc);
		s.animation().load(desc, version);
	    }
	}
    }

    public int seriko_version() {
	try {
	    int version = 0;
	    
	    Map desc = (Map)surfaces_txt.get("descript");
	    if (desc != null) {
		String ver = (String)desc.get("version");
		version = Integer.parseInt(ver);
	    }

	    if (version != 0 && version != 1) {
		throw new Exception("Invalid SERIKO version: "+version);
	    }
	    return version;
	}
	catch (Exception e) {
	    throw new RuntimeException(
		"Failed to see SERIKO version.", e);
	}
    }
}