package koropeta.gui.surfaceeditor.animation;
import koropeta.core.*;
import koropeta.core.animation.*;
import java.util.*;

public class Player {
    Shell shell;
    int base_surface_id;
    AnimationHandleComponent ahc;
    ThreadGroup thread_group;
    Map threads; // group id => PlayerThread
    List player_listeners;
    boolean in_exclusive;

    public Player(Shell shell, int base_surface_id, AnimationHandleComponent ahc) {
	this.shell = shell;
	this.base_surface_id = base_surface_id;
	this.ahc = ahc;
	this.thread_group = new ThreadGroup("Player Threads");
	this.threads = new Hashtable();
	this.player_listeners = new Vector();
    }

    public void addPlayerListener(PlayerListener listener) {
	if (!player_listeners.contains(listener)) {
	    player_listeners.add(listener);
	}
    }

    public void removePlayerListener(PlayerListener listener) {
	player_listeners.remove(listener);
    }

    public void play_all() {
	// ahcを初期状態に戻す
	shadow_play(-1, -1);

	emit_event(new PlayerEvent(Player.this, PlayerEvent.BEGAN_ALL));
	
	for (Iterator ite = ahc.animation().iterator(); ite.hasNext(); ) {
	    AnimGroup group = (AnimGroup)ite.next();
	    if (!group.is_autostarted()) {
		continue;
	    }
	    
	    PlayerThread thread = new PlayerThread(
		group, false /* no start */, false /* cleanup */, true /* periodic */);
	    threads.put(new Integer(group.id()), thread);
	    thread.start();
	}

	cleanup_dead_threads();
    }

    public void play_group(int group_id) {
	AnimGroup group = ahc.animation().group(group_id);

	// ahcをこのグループの初期状態に戻す。
	shadow_play(group_id, -1);

	emit_event(new PlayerEvent(Player.this, PlayerEvent.BEGAN_ALL));

	PlayerThread thread = new PlayerThread(
	    group, true /* no start */, true /* cleanup */, false /* periodic */);
	threads.put(new Integer(group_id), thread);
	thread.start();

	cleanup_dead_threads();
    }

    public void stop_group(int group_id) {
	synchronized (threads) {
	    Integer id = new Integer(group_id);
	    PlayerThread thread = (PlayerThread)threads.get(id);
	    
	    if (thread != null) {
		thread.terminate();
		threads.remove(id);
	    }
	}
    }

    public boolean is_playing(int group_id) {
	PlayerThread thread = (PlayerThread)threads.get(new Integer(group_id));
	return thread != null && thread.isAlive();
    }
    
    public boolean is_playing() {
	return count_active_threads() > 0;
    }

    public void cleanup_dead_threads() {
	synchronized (threads) {
	    for (Iterator ite = threads.entrySet().iterator(); ite.hasNext(); ) {
		Map.Entry entry = (Map.Entry)ite.next();
		PlayerThread thread = (PlayerThread)entry.getValue();

		if (!thread.isAlive()) {
		    ite.remove();
		}
	    }
	}
    }

    public int count_active_threads() {
	Thread[] array = new Thread[thread_group.activeCount() + 10];
	thread_group.enumerate(array);

	int count = 0;
	for (int i = 0; i < array.length; i++) {
	    if (array[i] != null && array[i] instanceof PlayerThread) {
		count++;
	    }
	}
	return count;
    }

    public void stop_all() {
	terminate_threads();
    }
    protected void terminate_threads() {
	synchronized (threads) {
	    for (Iterator ite = threads.values().iterator(); ite.hasNext(); ) {
		PlayerThread thread = (PlayerThread)ite.next();
		
		thread.terminate();
	    }
	    threads.clear();
	}
    }

    protected synchronized void acquire_exclusive_lock() {
	in_exclusive = true;
	
	synchronized (threads) {
	    for (Iterator ite = threads.values().iterator(); ite.hasNext(); ) {
		PlayerThread thread = (PlayerThread)ite.next();
		
		if (Thread.currentThread() != thread) {
		    thread.interrupt(); // 實行中のアニメーションを早く終わらせる
		    // FIXME: 嚴密には終了を待った方が良いが、そこまでする必要はあるかどうか。
		}
	    }
	}
    }

    public void shadow_play(int group_id, int pattern_id) {
	// 指定したグループの指定したパターンの状態をahcに適用する。
	// group_idまたはpattern_idが負の値なら、初期状態に戻す。
	shadow_play(group_id, pattern_id, true);
    }

    public void shadow_play(int group_id, int pattern_id, boolean cleanup_all) {
	AnimGroup group = null;
	AnimPattern pattern = null;
	if (group_id >= 0) {
	    group = ahc.animation().group(group_id);
	    if (pattern_id >= 0) {
		pattern = group.get(pattern_id);
	    }
	}

	if (cleanup_all) {
	    // 先づは關係無いグループのレイヤを消す。
	    terminate_threads();
	    if (group_id >= 0) {
		ahc.remove_group_layers_except(group_id);
	    }
	    else {
		ahc.remove_group_layers();
	    }
	}

	{ // moveを檢索
	    int base_x = 0, base_y = 0;
	    for (int i = pattern_id; i >= 0; i--) {
		AnimPattern pat = group.get(i);
		
		if (pat.method().equals("move")) {
		    base_x = pat.offset_x();
		    base_y = pat.offset_y();
		    break;
		}
	    }
	    ahc.offset(base_x, base_y);
	}

	{ // baseを檢索
	    int surface_id = base_surface_id;
	    for (int i = pattern_id; i >= 0; i--) {
		AnimPattern pat = group.get(i);
		
		if (pat.method().equals("base")) {
		    surface_id = pat.surface_id();
		    break;
		}
	    }

	    Surface s = (Surface)shell.get_surfaces().get(new Integer(surface_id));
	    if (s == null) {
		// これは許されないのでfallback
		// FIXME: fallbackした事をユーザへ何如にして通知するか。
		s = (Surface)shell.get_surfaces().get(new Integer(base_surface_id));
	    }
	    ahc.base_surface(s.image());
	}

	{ // グループレイヤを作るものを檢索
	    int surface_id = -1;
	    int x = 0, y = 0;
	    for (int i = pattern_id; i >= 0; i--) {
		AnimPattern pat = group.get(i);
		
		if (pat.does_make_group_layer()) {
		    surface_id = pat.surface_id();
		    x = pat.offset_x();
		    y = pat.offset_y();
		    break;
		}
	    }
	    
	    Surface s = (Surface)shell.get_surfaces().get(new Integer(surface_id));
	    if (s == null) {
		// レイヤを消去
		ahc.group_layer(group_id, null, 0, 0);
	    }
	    else {
		// レイヤを表示
		ahc.group_layer(group_id, s.image(), x, y);
	    }
	}

	// 今囘のパターンがmoveなら基本畫像を、レイヤ系ならレイヤを、
	// それぞれアクティブにする。
	if (pattern != null && pattern.method().equals("move")) {
	    ahc.activate_base_surface();
	}
	else if (pattern != null && pattern.does_make_group_layer()) {
	    ahc.activate_group_layer(group_id);
	}
	else {
	    ahc.active(-1);
	}
    }

    protected void emit_event(final PlayerEvent ev) {
	Runnable runnable = new Runnable() {
		public void run() {
		    for (Iterator ite = player_listeners.iterator(); ite.hasNext(); ) {
			PlayerListener listener = (PlayerListener)ite.next();
			
			if (ev.type() == PlayerEvent.BEGAN_ALL) {
			    listener.playerBeganAll(ev);
			}
			else if (ev.type() == PlayerEvent.ENDED_ALL) {
			    listener.playerEndedAll(ev);
			}
			else if (ev.type() == PlayerEvent.BEGAN_GROUP) {
			    listener.playerBeganGroup(ev);
			}
			else if (ev.type() == PlayerEvent.ENDED_GROUP) {
			    listener.playerEndedGroup(ev);
			}
			else if (ev.type() == PlayerEvent.CHANGED_PATTERN) {
			    listener.playerChangedPattern(ev);
			}
			else {
			    throw new RuntimeException("Unknown type: "+ev.type());
			}
		    }
		}
	    };
	Thread t = new Thread(runnable);
	t.start();
    }

    protected class PlayerThread extends Thread {
	volatile boolean flag_term;
	AnimGroup group;
	boolean no_start;
	boolean cleanup;
	boolean periodic;

	public PlayerThread(
	    AnimGroup group, boolean no_start, boolean cleanup, boolean periodic) {

	    super(thread_group, "PlayerThread (Group: "+group.id()+")");
	    
	    this.group    = group;
	    this.no_start = no_start; // start, alternativestartメソッドを無視
	    this.cleanup  = cleanup;  // スレッド終了時にahcを完全に初期状態に戻す
	    this.periodic = periodic; // 周期的に實行し續ける。但しsometimes, rarely, random, alwaysの場合のみ

	    if (!group.is_periodic()) {
		this.periodic = false;
	    }

	    setPriority(Thread.MIN_PRIORITY);
	}

	public final boolean periodic() {
	    return periodic;
	}
	
	public void run() {
	    emit_event(new PlayerEvent(Player.this, PlayerEvent.BEGAN_GROUP, group.id()));

	    final boolean is_exclusive =
		group.option() != null &&
		group.option().equals("exclusive");
	    
	    boolean played_once = false;
	    while (!flag_term) {
		if (periodic) {
		    // intervalに應じてスリープ
		    String interval = group.interval();
		    while (!flag_term) {
			double prob;
			
			if (interval.equals("sometimes")) {
			    prob = 1.0 / 2.0;
			}
			else if (interval.equals("rarely")) {
			    prob = 1.0 / 4.0;
			}
			else if (interval.equals("random")) {
			    if (group.interval_arg() <= 0) {
				prob = 0;
			    }
			    else {
				prob = 1.0 / group.interval_arg();
			    }
			}
			else if (interval.equals("always")) {
			    prob = 1.0;
			}
			else {
			    throw new RuntimeException("Unknown interval: "+interval);
			}

			if (Math.random() < prob) {
			    break;
			}
			else {
			    try {
				sleep(1000);
			    }
			    catch (InterruptedException e) {}
			}
		    }

		    synchronized (Player.this) {
			// exclusiveロック中ならキャンセル
			if (in_exclusive) {
			    continue;
			}
			
			// 必要に應じてexclusiveロックを掛ける。
			if (is_exclusive) {
			    in_exclusive = true;
			    thread_group.interrupt();
			}
		    }
		}
		else {
		    if (played_once) {
			break;
		    }
		}
		played_once = true;
		
		for (int i = 0; i < group.size(); i++) {
		    AnimPattern pattern = group.get(i);
		    String method = pattern.method();
		    
		    if (flag_term) {
			/* 終了指示が出てゐる */
			break;
		    }
		    else if (!is_exclusive && in_exclusive) {
			/* 他のスレッドがexclusiveロックを掛けた */
			ahc.group_layer(group.id(), null, 0, 0);
			break;
		    }

		    emit_event(new PlayerEvent(Player.this, PlayerEvent.CHANGED_PATTERN, group.id(), i));
		    
		    if (pattern.does_sleep()) {
			try {
			    sleep(pattern.interval());
			}
			catch (InterruptedException e) {
			    if (flag_term) {
				break;
			    }
			}
		    }
		    
		    if (method.equals("move")) {
			ahc.offset(pattern.offset_x(), pattern.offset_y());
		    }
		    else if (method.equals("base")) {
			Surface s = (Surface)shell.get_surfaces().get(new Integer(pattern.surface_id()));
			if (s == null) {
			    // これは許されないのでfallback
			    // FIXME: fallbackした事をユーザへ何如にして通知するか。
			    s = (Surface)shell.get_surfaces().get(new Integer(base_surface_id));
			}
			ahc.base_surface(s.image());
		    }
		    else if (pattern.does_make_group_layer()) {
			if (pattern.surface_id() == -1) {
			    // terminate this
			    ahc.group_layer(group.id(), null, 0, 0);
			    break;
			}
			else if (pattern.surface_id() == -2) {
			    // terminate all
			    ahc.group_layer(group.id(), null, 0, 0);
			    terminate_threads();
			    break;
			}
			else {
			    Surface s = (Surface)shell.get_surfaces().get(new Integer(pattern.surface_id()));
			    ahc.group_layer(
				group.id(),
				s == null ? null : s.image(),
				pattern.offset_x(),
				pattern.offset_y());
			}
		    }
		    else if (method.equals("start")) {
			if (!no_start) {
			    start(pattern.start_id());
			}
		    }
		    else if (method.equals("alternativestart")) {
			if (!no_start) {
			    Vector ids = pattern.alternative_ids();
			    if (ids != null && ids.size() > 0) {
				int id = (int)(Integer.MAX_VALUE * Math.random()) % ids.size();
				start(id);
			    }
			}
		    }
		    
		    if (!periodic && i == group.size() - 1) {
			// terminateされるまでスレッドを生かしたまま休止
			while (!flag_term) {
			    try {
				sleep(10000);
			    }
			    catch (InterruptedException e) {}
			}
		    }
		}

		if (is_exclusive) {
		    synchronized (Player.this) {
			in_exclusive = false;
		    }
		}
	    }

	    if (cleanup) {
		shadow_play(group.id(), -1, false);
	    }
	    emit_event(new PlayerEvent(Player.this, PlayerEvent.ENDED_GROUP, group.id()));

	    if (count_active_threads() == 1) {
		emit_event(new PlayerEvent(Player.this, PlayerEvent.ENDED_ALL));
	    }
	}

	private void start(int id) {
	    AnimGroup target = ahc.animation().group(id);
	    if (target != null) {
		PlayerThread old = (PlayerThread)threads.get(new Integer(target.id()));
		if (old != null) {
		    old.terminate();
		}
		
		PlayerThread thread = new PlayerThread(
		    target, false /* no start */, false /* cleanup */, true /* periodic */);
		threads.put(new Integer(target.id()), thread);
		thread.start();
	    }
	}

	public void terminate() {
	    if (!isAlive()) {
		return;
	    }
	    
	    flag_term = true;

	    if (this != Thread.currentThread()) {
		while (true) {
		    interrupt();
		    
		    try {
			join();
			break;
		    }
		    catch (InterruptedException e) {}
		}
	    }
	    else {
		// 自分自身をjoinする事は出來ない。
		interrupt();
	    }
	}
    }
}